import os
import math
import time

import numpy as np
from scipy.sparse import csr_matrix

import ctypes
from numpy.ctypeslib import ndpointer

import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import SparseTensorValue as STV

from utilities import predict, vectorize, inverse_vectorize, calc_norm, zero_tensor
from utilities import Y as Yclass

import pdb

class Minibatch(object):
    def __init__(self, config, sess, network):
        """
        initialize operations and vairables that will be used in newton
        args:
            sess: tensorflow session
            outputs: output of the neural network (pre-softmax layer)
            loss: function to calculate loss
       """
        super(Minibatch, self).__init__()
        self.network = network
        self.u, self.v, self.p, self.q = network
        self.sess = sess

        lib = ctypes.cdll.LoadLibrary("./c_funs/cfuns.so")

        self.cpred_val = lib.cpred_val
        self.cpred_val.restype = ctypes.c_double
        self.cpred_val.argtypes = [
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ctypes.c_size_t,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_double,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ]

        self.cBQ = lib.cBQ
        self.cBQ.restype = None
        self.cBQ.argtypes = [
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ctypes.c_size_t,
                ctypes.c_size_t,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_double,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ]


        self.config = config

        #TODO define scope name
        self.param = tf.trainable_variables()

        self.param_u = tf.trainable_variables(scope="u_model")
        self.param_v = tf.trainable_variables(scope="v_model")

        FLOAT = tf.float64
        self.FLOAT = FLOAT
        self.NP_FLOAT = np.float64

        model_weight_u = vectorize(self.param_u)
        model_weight_v = vectorize(self.param_v)

        self.p_loss = 0
        self.p_loss_holder = tf.placeholder(FLOAT)
        self.f = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.nloss = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.reg = tf.Variable(0., dtype=FLOAT, trainable=False)

        self.adagrad_init_accum_value = 0.1
        self.adagrad_eps = 1e-7

        self.tic = 0
        self.total_running_time = 0
        self.nr_print = 0

        self.amp = 1.0
        self.amp *= self.config.m / self.config.bsize_i
        self.amp *= self.config.n / self.config.bsize_j

        # initial variables cached in CPU
        self.P = np.zeros([self.config.m, self.config.k], dtype=self.NP_FLOAT)
        self.Q = np.zeros([self.config.n, self.config.k], dtype=self.NP_FLOAT)

        self.Pt = np.ones([self.config.m, self.config.k1], dtype=self.NP_FLOAT)*self.config.r*math.sqrt(1/self.config.k1)
        self.Qt = np.ones([self.config.n, self.config.k1], dtype=self.NP_FLOAT)*math.sqrt(1/self.config.k1)

        # todo: try tf.linalg.tensor_diag
        self.R = np.ones([self.config.m], dtype=self.NP_FLOAT)
        self.S = np.ones([self.config.n], dtype=self.NP_FLOAT)

        RP = self.R[:, np.newaxis]*self.Pt
        SQ = self.S[:, np.newaxis]*self.Qt
        tPc = np.zeros([self.config.k1, self.config.k1], dtype=self.NP_FLOAT)
        tQc = np.zeros([self.config.k1, self.config.k1], dtype=self.NP_FLOAT)
        num_segment = math.ceil(self.config.m/self.config.bsize)
        for i in range(num_segment):
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.m))
            tPc += RP[idx].T.dot(RP[idx])
        num_segment = math.ceil(self.config.n/self.config.bsize)
        for i in range(num_segment):
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.n))
            tQc += SQ[idx].T.dot(SQ[idx])
        self.tPC_tQC_F = (tPc * tQc).sum()


        self.R_holder = tf.placeholder(dtype=self.FLOAT)
        self.S_holder = tf.placeholder(dtype=self.FLOAT)

        # initial Grammian variabels in GPU
        kk_shape = tf.zeros(shape=[self.config.k, self.config.k], dtype=FLOAT)
        self.Pc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)
        self.Qc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)

        k1k_shape = tf.zeros(shape=[self.config.k1, self.config.k], dtype=FLOAT)
        self.hPc = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)
        self.hQc = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)

        zeros_u = tf.zeros(model_weight_u.get_shape(), dtype=FLOAT)
        zeros_v = tf.zeros(model_weight_v.get_shape(), dtype=FLOAT)

        # initial variable used in gradient
        self.g_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.g_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.ada_u = tf.Variable(zeros_u+self.adagrad_init_accum_value, dtype=FLOAT, trainable=False)
        self.ada_v = tf.Variable(zeros_v+self.adagrad_init_accum_value, dtype=FLOAT, trainable=False)

        self.d_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.d_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.s_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.s_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self._lambda = tf.placeholder(FLOAT, shape=[])

        self.update_model = self._update_model()
        self.gnorm = calc_norm(tf.concat([self.g_u, self.g_v], axis=0))

        self.cal_nloss = self._cal_nloss()
        self.cal_reg = self._cal_reg()
        self.cal_fun_val = self._cal_fun_val()

        self.tp = tf.placeholder(dtype=self.FLOAT)
        self.tq = tf.placeholder(dtype=self.FLOAT)

        self.zero_Pc_hPc = self._zero_gramian(self.Pc, self.hPc)
        self.zero_Qc_hQc = self._zero_gramian(self.Qc, self.hQc)

        self.cal_gramian_P = self._cal_gramian(self.p, self.tp, self.Pc, self.hPc, self.R_holder)
        self.cal_gramian_Q = self._cal_gramian(self.q, self.tq, self.Qc, self.hQc, self.S_holder)
        self.holder_bq = tf.placeholder(dtype=self.FLOAT)
        self.holder_bp = tf.placeholder(dtype=self.FLOAT)

        self.holder_p = tf.placeholder(self.FLOAT)
        self.holder_q = tf.placeholder(self.FLOAT)

        self.zero_g_u = zero_tensor(self.g_u)
        self.zero_g_v = zero_tensor(self.g_v)

        self.batch_grad_u = self._cal_batch_grad(self.holder_bq, self.holder_p, self.tp,
                self.Qc, self.hQc, self.p, self.param_u, self.g_u, self.R_holder)
        self.batch_grad_v = self._cal_batch_grad(self.holder_bp, self.holder_q, self.tq,
                self.Pc, self.hPc, self.q, self.param_v, self.g_v, self.S_holder)

        self.reg_grad_u = self._add_reg_grad(self.param_u, self.g_u)
        self.reg_grad_v = self._add_reg_grad(self.param_v, self.g_v)


        self.update_adagrad = tf.group(*[
            tf.assign(self.ada_u, self.ada_u + self.g_u*self.g_u),
            tf.assign(self.ada_v, self.ada_v + self.g_v*self.g_v)
        ])

        self.gradient_direction = tf.group(*[
            tf.assign(self.s_u, -self.g_u),
            tf.assign(self.s_v, -self.g_v)
            ])

        self.adagrad_direction = tf.group(*[
            tf.assign(self.s_u, -self.g_u*tf.math.rsqrt(self.ada_u+self.adagrad_eps)),
            tf.assign(self.s_v, -self.g_v*tf.math.rsqrt(self.ada_v+self.adagrad_eps))
        ])

    def profile(self, *args, **kwargs):
        if self.config.profile:
            print(*args, **kwargs)


    def cal_pred_val_and_ploss(self, Y, P, Q, R, S, idx_i, idx_j):
        """
        Calculate L^+ part's LogLoss via expyy values of nonzeros in Y
        """
        batch_y = Y.Label[idx_i, :][:, idx_j]
        y_imp = Y.Y_imp[idx_i, :][:, idx_j].data

        indptr = batch_y.indptr
        indices = batch_y.indices
        m = batch_y.shape[0]

        omega = self.config.omega

        y_hat = np.zeros_like(y_imp)
        ysigma = np.zeros_like(y_imp)

        self.p_loss = self.cpred_val(P, Q, self.config.k, indptr, indices, m, R, S, omega,
                batch_y.data, y_imp, y_hat, ysigma)

        return ysigma, y_hat, y_imp


    def cal_ival(self, Y):
        coos = Y.Y_imp.nonzero()
        ps = self.Pt[coos[0]]
        qs = self.Qt[coos[1]]
        Y.Y_imp.data = (ps*qs).sum(1)

    def transpose(self, Y):
        YT = Yclass(Y.n, Y.m, Y.nnzs)

        nonzero = Y.Label.transpose().nonzero()
        YT.Label = csr_matrix((np.array(Y.Label.transpose()[nonzero]).squeeze(), nonzero), shape=(Y.n, Y.m))
        
        YT.Y_imp = csr_matrix((np.array(Y.Y_imp.transpose()[nonzero]).squeeze(), nonzero), shape=(Y.n, Y.m))
        
        return YT

    def _cal_gramian(self, p, tp, Pc, hPc, R_holder):
        RP = tf.multiply(R_holder[:, tf.newaxis], p)
        update_Pc = tf.assign(Pc, Pc + tf.transpose(p) @ RP)
        update_hPc = tf.assign(hPc, hPc + tf.transpose(tp) @ RP)
        return tf.group(*[update_Pc, update_hPc])
    
    def _zero_gramian(self, Pc, hPc):
        zero_Pc = zero_tensor(Pc)
        zero_hPc = zero_tensor(hPc)
        return tf.group(*[zero_Pc, zero_hPc])

    def _cal_nloss(self):
        PcQc_F = tf.reduce_sum(tf.math.multiply(self.Pc, self.Qc), [0,1])
        hPc_hQc_F = tf.reduce_sum(tf.math.multiply(self.hPc, self.hQc), [0,1])
        # todo: add tPc_tQc_F to nloss?
        return tf.assign(self.nloss, self.tPC_tQC_F/2 + PcQc_F/2 - hPc_hQc_F)

    def _cal_reg(self):
        model_weight = vectorize(self.param)
        reg = (calc_norm(model_weight))**2
        return tf.assign(self.reg, reg*self.config._lambda/2)

    def microbatches(self, input):
        m = input.shape[0]
        num_segment = math.ceil(m/self.config.bsize)
        for i in range(num_segment):
            load_time = time.time()
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, m))
            start, end = idx[0], idx[-1]+1
            subset = slice(start, end)
            batch_input = input[subset]
            batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
            self.config.elapsed_time += time.time() - load_time
            
            yield subset, batch_input


    def cal_output(self, zero_gramian, cal_gramian, input, place_holder_u, tp, p, tP, P, R_holder, R, idx_i):
        """
        A function to evaluate either function value, global gradient or sub-sampled Gv
        """

        self.sess.run(zero_gramian)

        u = place_holder_u
        for subset, batch_input in self.microbatches(input):
            P[idx_i[subset]],_ = self.sess.run([p, cal_gramian], feed_dict={
                u: batch_input, tp: tP[subset], R_holder: R[subset]})

    def _cal_batch_grad(self, holder_bq, holder_p, holder_tp, Qc, hQc, p, param, g, R_holder):
        pre_grad_neg = tf.multiply(R_holder[:, tf.newaxis], holder_p @ Qc - holder_tp @ hQc)
        pre_grad = holder_bq + self.config.omega*pre_grad_neg
        grad = tf.gradients(ys=p, xs=param, grad_ys=pre_grad)
        grad = vectorize(grad)
        update_grad = tf.assign(g, g+grad*self.amp)
        return update_grad

    def _add_reg_grad(self, param, g):
        model_weight = vectorize(param)
        reg_grad = model_weight*self.config._lambda
        return tf.assign(g, g+reg_grad)

    def print_info(self, k, i, num_segment, Y, YT, Utr, V, val_batch, log_file):
        toc = time.time()
        epoch = k + i/num_segment
        self.total_running_time += toc-self.tic
        self.fun_val(Utr, V, Y, YT, self.network)
        f = self.sess.run(self.f)
        output_str = 'iter: {:.2f} f: {:.3f} time: {:.3f}'.format(epoch, f, self.total_running_time)
        print(output_str)
        print(output_str, file=log_file)
        if val_batch is not None and (self.nr_print+1) % 5 == 0:
            Ute, Yte = val_batch
            psprecision, ndcg = predict(self.sess, self.network, Yte.Label, Ute, V, self.config)
            precision = "p@k"+ " " + " ".join([str(round(i*100,3)) for i in psprecision])
            ndcg = "ndcg@k" + " " + " ".join([str(round(i*100,3)) for i in ndcg])
            print(precision + "\t" + ndcg)
            print(precision + "\t" + ndcg, file=log_file)
        log_file.flush()
        self.tic = time.time()
        self.nr_print += 1
        
    def cal_BQ(self, Y, P, Q, R, S, idx_i, idx_j):
        ysigma, y_hat, y_imp = self.cal_pred_val_and_ploss(Y, P, Q, R, S, idx_i, idx_j)

        label_ = Y.Label[idx_i, :][:, idx_j]
        indptr = label_.indptr
        indices = label_.indices

        label_sub = label_.data

        m_sub = label_.shape[0]

        omega = self.config.omega

        BQ = np.zeros((m_sub, self.config.k), dtype = self.NP_FLOAT)
        self.cBQ(Q, self.config.k, indptr, indices, m_sub, 0, R, S, omega,
                label_sub, y_imp, y_hat, ysigma, BQ)
        return BQ

    def cal_grad(self, zero_g, batch_grad, reg_grad, input, Y, P, Q, tP, place_holder_u, p, holder_bq, holder_p, holder_tp, R_holder, R, S, idx_i, idx_j):
        # before estimation starts, need to zero out grad
        self.sess.run(zero_g)

        u = place_holder_u
        bP = np.ascontiguousarray(P)
        btP = np.ascontiguousarray(tP)

        s = time.time()
        bBQ =  self.cal_BQ(Y, P, Q, R, S, idx_i, idx_j)
        self.profile("bBQ:",time.time()-s)
        s = time.time()
        batch_input = STV(np.array(input.nonzero()).transpose(), input.data, input.shape)
        self.sess.run(batch_grad, feed_dict={u: batch_input, holder_bq: bBQ,
                                             holder_p: bP, holder_tp: btP,
                                             R_holder: R})
        self.profile("gpu_grad:",time.time()-s)
        
        self.sess.run(reg_grad)


    def _update_model(self):
        update_model_ops = []

        x = inverse_vectorize(self.s_u, self.param_u)
        for i, p in enumerate(self.param_u):
            op = tf.assign(p, p + self.config.lr * x[i])
            update_model_ops.append(op)

        x = inverse_vectorize(self.s_v, self.param_v)
        for i, p in enumerate(self.param_v):
            op = tf.assign(p, p + self.config.lr * x[i])
            update_model_ops.append(op)
        return tf.group(*update_model_ops)

    def _cal_fun_val(self):
        return tf.assign(self.f, self.config.omega*self.cal_nloss + self.p_loss_holder + self.cal_reg)


    def fun_val(self, Utr, V, Y, YT, network):
        idx_i = np.arange(self.config.m)
        idx_j = np.arange(self.config.n)
        u,v,p,q = network
        s = time.time()
        self.cal_output(self.zero_Pc_hPc, self.cal_gramian_P,  Utr, u, self.tp, p, self.Pt, self.P, self.R_holder, self.R, idx_i)
        self.cal_output(self.zero_Qc_hQc, self.cal_gramian_Q, V, v, self.tq, q, self.Qt, self.Q, self.S_holder, self.S, idx_j)
        self.profile("output:",time.time()-s)
        
        s = time.time()
        self.cal_pred_val_and_ploss(Y, self.P, self.Q, self.R, self.S, idx_i, idx_j)
        self.cal_pred_val_and_ploss(YT, self.Q, self.P, self.S, self.R, idx_j, idx_i)
        self.profile("ploss:",time.time()-s)

        s = time.time()
        self.sess.run(self.cal_fun_val, feed_dict={self.p_loss_holder:self.p_loss})
        self.profile("nloss:",time.time()-s)

    def minibatch(self, Y, Utr, V, val_batch, saver):
        """
        Conduct newton steps for training
        args:
            full_batch & val_batch: provide training set and validation set. The function will
                save the best model evaluted on validation set for future prediction.
            network: a tuple contains (x, y, loss, outputs).
            test_network: a tuple similar to argument network. If you use layers which behave differently
                in tesse such as batchnorm, a separate test_network is needed.
        return:
            None
        """
        # check whether data is valid
        assert (Y.m == Utr.shape[0]) and (Y.n == V.shape[0])


        tf.summary.scalar('loss', self.f)
        merged = tf.summary.merge_all()
        train_writer = tf.summary.FileWriter('./summary/train', self.sess.graph)

        print(self.config.args)
        if not self.config.screen_log_only:
            log_file = open(self.config.log_file, 'w')
            print(self.config.args, file=log_file)
        
        self.cal_ival(Y)

        YT = self.transpose(Y)

        self.fun_val(Utr, V, Y, YT, self.network)

        f = self.sess.run(self.f)
        output_str = 'initial f: {:.3f}'.format(f)
        print(output_str)

        if not self.config.screen_log_only:
            print(output_str, file=log_file)

        best_acc = 0.0

        if self.config.lr == 0:
            raise ValueError('Learning rate  cannot be zero!')

        total_running_time = 0.0
        self.config.elapsed_time = 0.0
        u, v, p, q = self.network
        self.tic = time.time()
        for k in range(self.config.iter_max):

            nb_i = math.ceil(self.config.m / self.config.bsize_i)
            nb_j = math.ceil(self.config.n / self.config.bsize_j)
            num_segment = nb_i * nb_j
            for b in range(num_segment):

                idx_i = np.random.choice(np.arange(0, self.config.m),
                        size=self.config.bsize_i, replace=False)

                idx_j = np.random.choice(np.arange(0, self.config.n),
                        size=self.config.bsize_j, replace=False)

                s = time.time()
                self.cal_output(self.zero_Pc_hPc, self.cal_gramian_P,  Utr[idx_i], u, self.tp, p, self.Pt[idx_i], self.P, self.R_holder, self.R[idx_i], idx_i)
                self.cal_output(self.zero_Qc_hQc, self.cal_gramian_Q, V[idx_j], v, self.tq, q, self.Qt[idx_j], self.Q, self.S_holder, self.S[idx_j], idx_j)
                self.profile("output:",time.time()-s)
 
                self.cal_grad(self.zero_g_u, self.batch_grad_u, self.reg_grad_u,
                    Utr[idx_i], Y, self.P[idx_i], self.Q[idx_j], self.Pt[idx_i], u, p, self.holder_bq, self.holder_p, self.tp, self.R_holder, self.R[idx_i], self.S[idx_j], idx_i, idx_j)

                self.cal_grad(self.zero_g_v, self.batch_grad_v, self.reg_grad_v,
                    V[idx_j], YT, self.Q[idx_j], self.P[idx_i], self.Qt[idx_j], v, q, self.holder_bp, self.holder_q, self.tq, self.S_holder, self.S[idx_j], self.R[idx_i], idx_j, idx_i)

                if self.config.adagrad:
                    self.sess.run(self.update_adagrad)
                    self.sess.run(self.adagrad_direction)
                else:
                    self.sess.run(self.gradient_direction)
                self.sess.run(self.update_model)

                if b % max(1, round(num_segment/10)) == 0: self.print_info(k, b, num_segment, Y, YT, Utr, V, val_batch, log_file)


        if val_batch is None:
            checkpoint_path = self.config.model_file
            save_path = saver.save(self.sess, checkpoint_path)
            output_str = 'Model at the last iteration saved in {}\r\n'.format(save_path)
            output_str += 'total running time {:.3f}s'.format(self.total_running_time)
        else:
            output_str = 'psprecision: {} | ndcg: {} | total running time {:.3f}s'.\
                format(psprecision, ndcg, total_running_time)
        print(output_str)
        if not self.config.screen_log_only:
            print(output_str, file=log_file)
            log_file.close()
