import os
import math
import time

import numpy as np
from scipy.sparse import csr_matrix

import ctypes
from numpy.ctypeslib import ndpointer

import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import SparseTensorValue as STV

from utilities import predict, vectorize, inverse_vectorize, calc_norm, zero_tensor
from utilities import Y as Yclass
import pdb



class Sogram(object):
    def __init__(self, config, sess, network):
        """
        initialize operations and vairables that will be used in newton
        args:
            sess: tensorflow session
            outputs: output of the neural network (pre-softmax layer)
            loss: function to calculate loss
       """
        super(Sogram, self).__init__()
        self.network = network
        self.u, self.v, self.p, self.q = network
        self.sess = sess

        lib = ctypes.cdll.LoadLibrary("./c_funs/cfuns.so")

        self.cpred_val = lib.cpred_val_sogram
        self.cpred_val.restype = ctypes.c_double
        self.cpred_val.argtypes = [
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ctypes.c_int,
                ctypes.c_size_t,
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_int32, flags="C_CONTIGUOUS"),
                ctypes.c_double,
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                ]


        self.config = config
        self.param = tf.trainable_variables()

        self.param_u = tf.trainable_variables(scope="u_model")
        self.param_v = tf.trainable_variables(scope="v_model")

        FLOAT = tf.float64
        self.FLOAT = FLOAT
        self.NP_FLOAT = np.float64


        model_weight_u = vectorize(self.param_u)
        model_weight_v = vectorize(self.param_v)

        self.p_loss = 0
        self.p_loss_holder = tf.placeholder(FLOAT)
        self.f = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.nloss = tf.Variable(0., dtype=FLOAT, trainable=False)
        self.reg = tf.Variable(0., dtype=FLOAT, trainable=False)

        self.adagrad_init_accum_value = 0.1
        self.adagrad_eps = 1e-7

        self.tic = 0
        self.total_running_time = 0
        self.nr_print = 0




        # initial variables cached in CPU
        self.P = np.zeros([self.config.sogram_bsize, self.config.k], dtype=self.NP_FLOAT)
        self.Q = np.zeros([self.config.sogram_bsize, self.config.k], dtype=self.NP_FLOAT)

        self.P_large = np.zeros([self.config.bsize, self.config.k], dtype=self.NP_FLOAT)
        self.Q_large = np.zeros([self.config.bsize, self.config.k], dtype=self.NP_FLOAT)

        self.Pt = np.ones([self.config.m, self.config.k1], dtype=self.NP_FLOAT)*self.config.r*math.sqrt(1/self.config.k1)
        self.Qt = np.ones([self.config.n, self.config.k1], dtype=self.NP_FLOAT)*math.sqrt(1/self.config.k1)

        # todo: try tf.linalg.tensor_diag
        self.R = np.zeros([self.config.m], dtype=self.NP_FLOAT)
        self.S = np.zeros([self.config.n], dtype=self.NP_FLOAT)

        # calculate tPc, tQc
        RP = self.Pt
        SQ = self.Qt
        tPc = np.zeros([self.config.k1, self.config.k1], dtype=self.NP_FLOAT)
        tQc = np.zeros([self.config.k1, self.config.k1], dtype=self.NP_FLOAT)
        num_segment = math.ceil(self.config.m/self.config.bsize)
        for i in range(num_segment):
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.m))
            tPc += RP[idx].T.dot(RP[idx])
        num_segment = math.ceil(self.config.n/self.config.bsize)
        for i in range(num_segment):
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.n))
            tQc += SQ[idx].T.dot(SQ[idx])
        self.tPC_tQC_F = (tPc * tQc).sum()

        self.R_holder = tf.placeholder(dtype=self.FLOAT)
        self.S_holder = tf.placeholder(dtype=self.FLOAT)


        # initial Grammian variabels in GPU
        kk_shape = tf.zeros(shape=[self.config.k, self.config.k], dtype=FLOAT)
        self.Pc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)
        self.Qc = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)

        self.Pcp = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)
        self.Qcp = tf.Variable(kk_shape, dtype=FLOAT, trainable=False)


        k1k_shape = tf.zeros(shape=[self.config.k1, self.config.k], dtype=FLOAT)
        self.hPc = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)
        self.hQc = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)

        self.hPcp = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)
        self.hQcp = tf.Variable(k1k_shape, dtype=FLOAT, trainable=False)

        zeros_u = tf.zeros(model_weight_u.get_shape(), dtype=FLOAT)
        zeros_v = tf.zeros(model_weight_v.get_shape(), dtype=FLOAT)

        # initial variable used in gradient
        self.g_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.g_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self.ada_u = tf.Variable(zeros_u+self.adagrad_init_accum_value, dtype=FLOAT, trainable=False)
        self.ada_v = tf.Variable(zeros_v+self.adagrad_init_accum_value, dtype=FLOAT, trainable=False)

        self.s_u = tf.Variable(zeros_u, dtype=FLOAT, trainable=False)
        self.s_v = tf.Variable(zeros_v, dtype=FLOAT, trainable=False)

        self._lambda = tf.placeholder(FLOAT, shape=[])

        self.update_model = self._update_model()
        self.gnorm = calc_norm(tf.concat([self.g_u, self.g_v], axis=0))


        self.tp = tf.placeholder(dtype=self.FLOAT)
        self.tq = tf.placeholder(dtype=self.FLOAT)

        self.zero_Pc_hPc = self._zero_gramian(self.Pc, self.hPc)
        self.zero_Qc_hQc = self._zero_gramian(self.Qc, self.hQc)

        self.cal_nloss = self._cal_nloss()
        self.cal_reg = self._cal_reg()
        self.cal_fun_val = self._cal_fun_val()

        self.zero_Pcp_hPcp = self._zero_gramian(self.Pcp, self.hPcp)
        self.zero_Qcp_hQcp = self._zero_gramian(self.Qcp, self.hQcp)

        self.cal_gramian_P = self._cal_gramian(self.p, self.tp, self.Pc, self.hPc, self.R_holder)
        self.cal_gramian_Q = self._cal_gramian(self.q, self.tq, self.Qc, self.hQc, self.S_holder)

        self.update_gramian_P = self._update_gramian(self.p, self.tp, self.Pcp, self.hPcp, self.R_holder)
        self.update_gramian_Q = self._update_gramian(self.q, self.tq, self.Qcp, self.hQcp, self.S_holder)

        self.holder_bq = tf.placeholder(dtype=self.FLOAT)
        self.holder_bp = tf.placeholder(dtype=self.FLOAT)

        self.holder_p = tf.placeholder(self.FLOAT)
        self.holder_q = tf.placeholder(self.FLOAT)

        self.zero_g_u = zero_tensor(self.g_u)
        self.zero_g_v = zero_tensor(self.g_v)

        self.batch_grad_u = self._cal_batch_grad(self.holder_bq, self.holder_p, self.tp,
                self.Qcp, self.hQcp, self.p, self.param_u, self.g_u, self.R_holder)
        self.batch_grad_v = self._cal_batch_grad(self.holder_bp, self.holder_q, self.tq,
                self.Pcp, self.hPcp, self.q, self.param_v, self.g_v, self.S_holder)

        self.reg_grad_u = self._add_reg_grad(self.param_u, self.g_u)
        self.reg_grad_v = self._add_reg_grad(self.param_v, self.g_v)


        self.update_adagrad = tf.group(*[
            tf.assign(self.ada_u, self.ada_u + self.g_u*self.g_u),
            tf.assign(self.ada_v, self.ada_v + self.g_v*self.g_v)
            ])

        self.gradient_direction = tf.group(*[
            tf.assign(self.s_u, -self.g_u),
            tf.assign(self.s_v, -self.g_v)
            ])

        self.adagrad_direction = tf.group(*[
            tf.assign(self.s_u, -self.g_u*tf.math.rsqrt(self.ada_u+self.adagrad_eps)),
            tf.assign(self.s_v, -self.g_v*tf.math.rsqrt(self.ada_v+self.adagrad_eps))
            ])


    def profile(self, *args, **kwargs):
        if self.config.profile:
            print(*args, **kwargs)


    def _cal_gramian(self, p, tp, Pc, hPc, R_holder):
        RP = tf.multiply(R_holder[:, tf.newaxis], p)
        update_Pc = tf.assign(Pc, Pc + tf.transpose(p) @ RP)
        update_hPc = tf.assign(hPc, hPc + tf.transpose(tp) @ RP)
        return tf.group(*[update_Pc, update_hPc])

    def _update_gramian(self, p, tp, Pc, hPc, R_holder):
        alpha = self.config.gramian_alpha
        amp = self.config.nnzs / self.config.sogram_bsize
        RP = tf.multiply(R_holder[:, tf.newaxis], p)
        update_Pc = tf.assign(Pc, (1-alpha)*Pc + alpha*amp*tf.transpose(p) @ RP)
        update_hPc = tf.assign(hPc, (1-alpha)*hPc + alpha*amp*tf.transpose(tp) @ RP)
        return tf.group(*[update_Pc, update_hPc])
    
    def _zero_gramian(self, Pc, hPc):
        zero_Pc = zero_tensor(Pc)
        zero_hPc = zero_tensor(hPc)
        return tf.group(*[zero_Pc, zero_hPc])

    def cal_output(self, input, idx, place_holder_u, tp, p, tP, P, R_holder, R):
        """
        A function to evaluate either function value, global gradient or sub-sampled Gv
        """

        u = place_holder_u

        #self.sess.run(zero_gramian)

        load_time = time.time()
        batch_input = input[idx]
        batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
        self.config.elapsed_time += time.time() - load_time

        P[:] = self.sess.run(p, feed_dict={u: batch_input})

    def _cal_batch_grad(self, holder_bq, holder_p, holder_tp, Qc, hQc, p, param, g, R_holder):
        pre_grad_neg = tf.multiply(R_holder[:, tf.newaxis], holder_p @ Qc - holder_tp @ hQc)
        pre_grad = holder_bq + self.config.omega*pre_grad_neg
        grad = tf.gradients(ys=p, xs=param, grad_ys=pre_grad)
        grad = vectorize(grad)
        amp = self.config.nnzs / self.config.sogram_bsize
        update_grad = tf.assign(g, g+grad*amp)
        return update_grad

    def _add_reg_grad(self, param, g):
        model_weight = vectorize(param)
        reg_grad = model_weight*self.config._lambda
        return tf.assign(g, g+reg_grad)


    def print_info(self, k, i, num_segment, Y, Utr, V, val_batch, log_file):
        toc = time.time()
        epoch = k + i/num_segment
        self.total_running_time += toc-self.tic
        self.fun_val(Y, Utr, V, self.network)
        f = self.sess.run(self.f)
        output_str = 'iter: {:.2f} f: {:.3f} time: {:.3f}'.format(epoch, f, self.total_running_time)
        print(output_str)
        print(output_str, file=log_file)
        if val_batch is not None and (self.nr_print+1) % 5 == 0:
            Ute, Yte = val_batch
            psprecision, ndcg = predict(self.sess, self.network, Yte.Label, Ute, V, self.config)
            precision = "p@k"+ " " + " ".join([str(round(i*100,3)) for i in psprecision])
            ndcg = "ndcg@k" + " " + " ".join([str(round(i*100,3)) for i in ndcg])
            print(precision + "\t" + ndcg)
            print(precision + "\t" + ndcg, file=log_file)
        log_file.flush()
        self.tic = time.time()
        self.nr_print += 1


    def cal_RS(self, Y):
        nzs = Y.Label.nonzero()
        for nz in nzs[0]:
            self.R[nz] += 1
        for nz in nzs[1]:
            self.S[nz] += 1
        self.R[self.R == 0] = 1
        self.S[self.S == 0] = 1
        self.R = 1/self.R
        self.S = 1/self.S


    def cal_grad(self, zero_g, batch_grad, reg_grad, input, bBQ, idx, P, Q, tP, place_holder_u, p, holder_bq, holder_p, holder_tp, R_holder, R, S):
        u = place_holder_u


        # before estimation starts, need to zero out grad
        self.sess.run(zero_g)

        load_time = time.time()
        batch_input = input[idx]
        batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
        self.config.elapsed_time += time.time() - load_time

        bP = np.ascontiguousarray(P)
        btP = np.ascontiguousarray(tP[idx])
        s = time.time()
        self.sess.run(batch_grad, feed_dict={u: batch_input, holder_bq: bBQ, holder_p: bP, holder_tp: btP, R_holder: R[idx]})
        self.profile("gpu_grad:",time.time()-s)
        self.sess.run(reg_grad)



    def _update_model(self):
        update_model_ops = []

        x = inverse_vectorize(self.s_u, self.param_u)
        for i, p in enumerate(self.param_u):
            op = tf.assign(p, p + self.config.lr * x[i])
            update_model_ops.append(op)

        x = inverse_vectorize(self.s_v, self.param_v)
        for i, p in enumerate(self.param_v):
            op = tf.assign(p, p + self.config.lr * x[i])
            update_model_ops.append(op)
        return tf.group(*update_model_ops)

    def _cal_nloss(self):
        PcQc_F = tf.reduce_sum(tf.math.multiply(self.Pc, self.Qc), [0,1])
        hPc_hQc_F = tf.reduce_sum(tf.math.multiply(self.hPc, self.hQc), [0,1])
        # todo: add tPc_tQc_F to nloss?
        return tf.assign(self.nloss, self.tPC_tQC_F/2 + PcQc_F/2 - hPc_hQc_F)

    def _cal_fun_val(self):
        return tf.assign(self.f, self.config.omega*self.cal_nloss + self.p_loss_holder + self.cal_reg)

    def _cal_reg(self):
        model_weight = vectorize(self.param)
        reg = (calc_norm(model_weight))**2
        return tf.assign(self.reg, reg*self.config._lambda/2)


    def update_gramian(self, update_gramian, input, idx, place_holder_u, tp, p, tP, P, R_holder, R):
        u = place_holder_u

        #self.sess.run(zero_gramian)

        load_time = time.time()
        batch_input = input[idx]
        batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
        self.config.elapsed_time += time.time() - load_time

        self.sess.run(update_gramian, feed_dict={
            u: batch_input, tp: tP[idx], R_holder: R[idx]})

    def cal_gramian(self, cal_gramian, input, idx, place_holder_u, tp, p, tP, P, R_holder, R):
        u = place_holder_u

        #self.sess.run(zero_gramian)

        load_time = time.time()
        batch_input = input[idx]
        batch_input = STV(np.array(batch_input.nonzero()).transpose(), batch_input.data, batch_input.shape)
        self.config.elapsed_time += time.time() - load_time

        P.fill(0)

        P[0:len(idx)],_ = self.sess.run([p, cal_gramian], feed_dict={
            u: batch_input, tp: tP[idx], R_holder: R[idx]})


    def grad(self, Utr, V, batch_y, coos, network):
        u,v,p,q = network
        u_idx = coos[0]
        v_idx = coos[1]

        s = time.time()
        self.cal_output(Utr, u_idx, u, self.tp, p, self.Pt, self.P, self.R_holder, self.R)
        self.cal_output(V, v_idx, v, self.tq, q, self.Qt, self.Q, self.S_holder, self.S)
        self.profile("output:",time.time()-s)

        nnz = len(u_idx)

        bp = np.zeros((nnz, self.config.k))
        bq = np.zeros((nnz, self.config.k))

        self.cpred_val(self.P, self.Q, self.config.k, self.Pt, self.Qt, self.config.k1,
                nnz, u_idx, v_idx, self.config.omega, batch_y, bp, bq)


        self.profile("pred:",time.time()-s)
        self.cal_grad(self.zero_g_u, self.batch_grad_u, self.reg_grad_u,
                Utr, bq, u_idx, self.P, self.Q, self.Pt, u, p, self.holder_bq, self.holder_p, self.tp, self.R_holder, self.R, self.S)

        self.cal_grad(self.zero_g_v, self.batch_grad_v, self.reg_grad_v,
                V, bp, v_idx, self.Q, self.P, self.Qt, v, q, self.holder_bp, self.holder_q, self.tq, self.S_holder, self.S, self.R)
        self.profile("grad:",time.time()-s)


    def fun_val_batch(self, Utr, V, batch_y, coos, network):
        u,v,p,q = network
        u_idx = coos[0]
        v_idx = coos[1]

        s = time.time()
        self.cal_gramian(self.cal_gramian_P, Utr, u_idx, u, self.tp, p, self.Pt, self.P_large, self.R_holder, self.R)
        self.cal_gramian(self.cal_gramian_Q, V, v_idx, v, self.tq, q, self.Qt, self.Q_large, self.S_holder, self.S)
        self.profile("output:",time.time()-s)
        
        s = time.time()

        nnz = len(u_idx)
        bp = np.zeros((nnz, self.config.k))
        bq = np.zeros((nnz, self.config.k))

        p_loss = self.cpred_val(self.P_large, self.Q_large, self.config.k, self.Pt, self.Qt, self.config.k1,
                nnz, u_idx, v_idx, self.config.omega, batch_y, bp, bq)
        self.p_loss += p_loss
        self.profile("pred:",time.time()-s)


    def fun_val(self, Y, Utr, V, network):
        self.sess.run([self.zero_Pc_hPc, self.zero_Qc_hQc])
        self.p_loss = 0

        num_segment = math.ceil(self.config.nnzs/self.config.bsize)
        for i in range(num_segment):

            load_time = time.time()
            idx = np.arange(i * self.config.bsize, min((i+1) * self.config.bsize, self.config.nnzs))
            start, end = idx[0],idx[-1]+1

            coos = [Y.Label.nonzero()[0][start:end], Y.Label.nonzero()[1][start:end]]
            batch_y = Y.Label.data[start:end]

            self.fun_val_batch(Utr, V, batch_y, coos, network)

        self.sess.run(self.cal_fun_val, feed_dict={self.p_loss_holder:self.p_loss})


    def sogram(self, Y, Utr, V, val_batch, saver):
        """
        Conduct newton steps for training
        args:
            full_batch & val_batch: provide training set and validation set. The function will
                save the best model evaluted on validation set for future prediction.
            network: a tuple contains (x, y, loss, outputs).
            test_network: a tuple similar to argument network. If you use layers which behave differently
                in tesse such as batchnorm, a separate test_network is needed.
        return:
            None
        """
        # check whether data is valid
        assert (Y.m == Utr.shape[0]) and (Y.n == V.shape[0])

        u,v,p,q = self.network


        print(self.config.args)
        if not self.config.screen_log_only:
            log_file = open(self.config.log_file, 'w')
            print(self.config.args, file=log_file)

        self.cal_RS(Y)

        self.fun_val(Y, Utr, V, self.network)
        f = self.sess.run(self.f)
        output_str = 'initial f: {:.3f}'.format(f)
        print(output_str)

        if not self.config.screen_log_only:
            print(output_str, file=log_file)


        self.config.elapsed_time = 0.0

        self.sess.run([self.zero_Pcp_hPcp, self.zero_Qcp_hQcp])

        self.tic = time.time()

        for k in range(self.config.iter_max):

            num_segment = math.ceil(self.config.nnzs/self.config.sogram_bsize) ** 2

            for  i in range(num_segment):

                #compute the estimation of Gramian matrices
                co_idx = np.random.choice(np.arange(0, self.config.nnzs),
                        size=self.config.sogram_bsize, replace=False)

                coos = [Y.Label.nonzero()[0][co_idx], Y.Label.nonzero()[1][co_idx]]
                batch_y = Y.Label.data[co_idx]

                self.update_gramian(self.update_gramian_P, Utr, coos[0],
                        u, self.tp, p, self.Pt, self.P, self.R_holder, self.R)

                self.update_gramian(self.update_gramian_Q, V, coos[1],
                        v, self.tq, q, self.Qt, self.Q, self.S_holder, self.S)

                #compute the gradient
                co_idx = np.random.choice(np.arange(0, self.config.nnzs),
                        size=self.config.sogram_bsize, replace=False)


                coos = [Y.Label.nonzero()[0][co_idx], Y.Label.nonzero()[1][co_idx]]
                batch_y = Y.Label.data[co_idx]

                self.grad(Utr, V, batch_y, coos, self.network)

                if self.config.adagrad:
                    self.sess.run(self.update_adagrad)
                    self.sess.run(self.adagrad_direction)
                else:
                    self.sess.run(self.gradient_direction)
                self.sess.run(self.update_model)

                if i % max(1, round(num_segment/10)) == 0: self.print_info(k, i, num_segment, Y, Utr, V, val_batch, log_file)


        if val_batch is None:
            checkpoint_path = self.config.model_file
            save_path = saver.save(self.sess, checkpoint_path)
            output_str = 'Model at the last iteration saved in {}\r\n'.format(save_path)
        else:
            output_str = 'psprecision: {} | ndcg: {} | total running time {:.3f}s'.\
                format(psprecision, ndcg, self.total_running_time)
        print(output_str)
        if not self.config.screen_log_only:
            print(output_str, file=log_file)
            log_file.close()
